/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* This CGI program edits the conflist file and does basic setup and does
 * basic creation and destruction of conference directories.
 *
 * It talks the following parameters:
 *
 *    op=     "C" - create a conference - both directory and conflist entry
 *            "M" - make conference - only directory
 *            "D" - delete a conference
 *    conf=   Name of conference (directory name) (all operations)
 *    loc=    Parent directory of conference (op=C only)
 *    dir=    loc/conf can be given if conf and dir are not.
 *    next=   URL of page to redirect to on success. (optional)
 *
 * On the create command, the conference directory defaults to BBS_DIR if
 * not given, and the alias list defaults to the conference name.  Aliases
 * in the alias list may be separated by most any kind of white space and
 * commas.
 */

#include "config.h"

#include <stdio.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <pwd.h>

#include "adm_conf.h"
#include "adm_cgi.h"
#include "udb_read.h"

char *getenv();

#if __STDC__ && HAVE_STDVARARGS
void die(const char *fmt,...)
#else
void die(fmt, va_alist)
    const char *fmt;
    va_dcl
#endif
{
#if HAVE_STDVARARGS
va_list ap;
#endif

    VA_START(ap,fmt);
    fputs("Content-Type: text/html\n\n",stdout);
    fputs("<TITLE>Conference Admininstration Error</TITLE>\n",stdout);
    fputs("<H1>Conference Administration Error</H1>\n",stdout);
    fputs("An error caused conference creation/deletion to fail:\n", stdout);
    fputs("<PRE>\n",stdout);
    VA_PRINTF(stdout, fmt, ap);
    fputs("</PRE>\n",stdout);
    VA_END(ap);
    exit(1);
}


/* KILLDIR - Remove a directory and all files below it. */

void killdir(char *dir)
{
pid_t cpid, wpid;

	if ((cpid= fork()) == 0)
	{
		execl("/bin/rm","rm","-rf",dir,(char *)NULL);
		exit(1);
	}

	while ((wpid= wait((int *)0)) != cpid && wpid != -1)
		;
}


/* DEL_PARTICIPATION - For all users, delete the named participation files.
 */

void del_participation(char *pfname)
{
#if defined(UD_PASSWD) && !defined(PASSWD_FILE)
	/* Not implemented yet for real unix accounts (because it is harder to
	 * do right)
	 */
#else
char *login;
struct passwd *pw;
char buf[BFSZ];

	/* Loop through all users */
	for (login= firstusr(); login != NULL; login= nextusr())
	{
		if ((pw= getdbnam(login)) == NULL) continue;
		sprintf(buf,"%s/%s",pw->pw_dir,pfname);
		unlink(buf);
	}
#endif
}


main(int argc, char **argv)
{
char *query, *err, *pfile;
char *op, *conf, *loc, *next, *lid, *dir;
char dbf[BFSZ],pfname[BFSZ];
#ifdef INSECURE
FILE *rfp;
#endif
char buf[BFSZ];

	/* Check that we are being run by httpd */
	if (geteuid() != CFADM_UID)
		die("Program lacks access to conferences -"
		    " suid bit not set? (EUID=%d)",geteuid());

#ifndef INSECURE
	/* Check that we are being run by httpd */
	if (getuid() != SERVER_UID)
		die("Program must be run by http daemon.");

	/* Check that we are authenticated as cfadm */
	if ((lid= getenv("REMOTE_USER")) == NULL || !inlist(lid,ADM_NAMES,","))
		die("You are not the conference administrator.");
#endif

	/* Load the query */
#ifdef INSECURE
	if (argc == 2 && !strcmp(argv[1],"-r"))
	{
		if ((rfp= fopen("/tmp/mkcfr","r")) != NULL)
		{
			fgets(buf, BFSZ, rfp);
			if ((query= strchr(buf,'\n')) != NULL) *query= '\0';
			query= buf;
			fclose(rfp);
		}
	}
	else
	{
#endif
        query= get_query();
#ifdef INSECURE
        if ((rfp= fopen("/tmp/mkcfr","w")) != NULL)
        {
		fputs(query, rfp);
		fputc('\n',rfp);
		fclose(rfp);
        }
    }
#endif
	load_query(query);

	/* Get the arguments */
	if ((op= getvar("op")) == NULL)
		die("No operation specified.");
	if (*op != 'C' && *op != 'D' && *op != 'M')
		die("Illegal operation %c specified.",*op);

	conf= getvar("conf");
	loc= getvar("loc");
	next= getvar("next");
	dir= getvar("dir");

	/* Make whichever of "conf", "loc" and "dir" are missing */
	if (conf == NULL || *conf == '\0' ||
	    (*op != 'D' && (loc == NULL || *loc == '\0')))
	{
		if (dir == NULL || *dir == '\0')
			die("Must specify conference name and location or directory.\n");
		/* Parse conference name and location out of full path */
		strcpy(dbf,dir);
		loc= dbf;
		do {
			conf= strrchr(dbf,'/');
			*(conf++)= '\0';
		} while (*conf == '\0');
	}
	else if (dir == NULL || *dir == '\0')
	{
		if (conf == NULL || *conf == '\0')
			die ("not enough information given "
			     "to identify conference.");
		sprintf(dbf,"%.100s/%.100s",loc,conf);
		dir= dbf;
	}

	switch (*op)
	{
	case 'C':
	case 'M':
		pfile= make_part_name(conf);
		if (err= check_conf(conf, loc, pfile))
			die("%s",err);

		make_confdir(dir, pfile, ADM_NAMES, 0, NULL);
		if (*op == 'M') break;
		add_conflist(conf,dir);
		break;

	case 'D':
		/* Delete all entries from conflist */
		del_conflist(dir);
		/* Get participation file name */
		readconfig(dir,pfname,NULL,NULL,NULL);
		/* Remove conference directory */
		killdir(dir);
		/* Nuke all user's participation files */
		del_participation(pfname);
		break;
	}

	/* If successful, redirect */
	printf("Content-Type: text/html\n");
	if (next != NULL)
		printf("Location: %s?conf=%s\n\n",next,conf);
	else if (*op == 'C' || *op == 'M')
		printf("\nConference %s created in directory %s\n",conf,loc);
	else
		printf("\nConference %s destroyed.\n",conf);
	exit(0);
}
